/*
 * Copyright (C) 1999-2025. Christian Heller.
 *
 * This file is part of the Cybernetics Oriented Interpreter (CYBOI).
 *
 * CYBOI is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License,
 * or (at your option) any later version.
 *
 * CYBOI is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with CYBOI. If not, see <http://www.gnu.org/licenses/>.
 *
 * Cybernetics Oriented Programming (CYBOP) <http://www.cybop.org/>
 * CYBOP Developers <cybop-developers@nongnu.org>
 *
 * @version CYBOP 0.28.0 2025-05-31
 * @author Christian Heller <christian.heller@cybop.org>
 */

#ifndef FLOW_LOGIC_CYBOL_FORMAT_CONSTANT_HEADER
#define FLOW_LOGIC_CYBOL_FORMAT_CONSTANT_HEADER

//
// System interface
//

#include <stddef.h> // wchar_t

//
// Library interface
//

#include "constant.h"

//
// Flow
//
// IANA media type: not defined
// Self-defined media type: flow
// This media type is a CYBOL extension.
//

/**
 * The flow/branch logic cybol format.
 *
 * Description:
 *
 * Branches the programme flow, depending on the criterion flag.
 *
 * Java example:
 *
 * if (flag != false) {
 *     System.out.println("true");
 * } else {
 *     System.out.println("false");
 * }
 *
 * Examples:
 *
 * <!-- Read models from file. -->
 * <node name="branch_voltage" channel="inline" format="flow/branch" model="">
 *     <node name="criterion" channel="inline" format="text/cybol-path" model=".settings.voltage_criterion"/>
 *     <node name="true" channel="file" format="element/part" model="indoor_climate/add_voltage.cybol"/>
 *     <node name="false" channel="file" format="element/part" model="indoor_climate/report_voltage_failure.cybol"/>
 * </node>
 *
 * <!-- Branch with one model. -->
 * <node name="if-then" channel="inline" format="flow/branch" model="">
 *     <node name="criterion" channel="inline" format="text/cybol-path" model=".app.var.flag_loose"/>
 *     <node name="true" channel="inline" format="text/cybol-path" model=".app.print_win"/>
 * </node>
 *
 * <!-- Branch with two models. -->
 * <node name="if-then-else" channel="inline" format="flow/branch" model="">
 *     <node name="criterion" channel="inline" format="text/cybol-path" model=".domain.flag"/>
 *     <node name="true" channel="inline" format="text/cybol-path" model=".domain.true_model"/>
 *     <node name="false" channel="inline" format="text/cybol-path" model=".domain.false_model"/>
 * </node>
 *
 * <!-- Verify existence of given url path. -->
 * <node name="test_query" channel="inline" format="access/indicate-exists" model="">
 *     <node name="result" channel="inline" format="text/cybol-path" model=".var.query_exists"/>
 *     <node name="part" channel="inline" format="text/cybol-path" model=".var.request:uri:query"/>
 * </node>
 * <node name="evaluate_query" channel="inline" format="flow/branch" model="">
 *     <node name="criterion" channel="inline" format="text/cybol-path" model=".var.query_exists"/>
 *     <node name="true" channel="inline" format="text/cybol-path" model=".logic.evaluate_query"/>
 *     <!--
 *         CAUTION! Sending a response is important, even if the query was empty or not understood,
 *         since some browsers request a "/favicon.ico" or other things and if no response was sent,
 *         the browser would wait forever and block requests from other clients.
 *     -->
 *     <node name="false" channel="inline" format="text/cybol-path" model=".logic.send.empty_message"/>
 * </node>
 *
 * <!-- Fill cell with default value if database value is empty. -->
 * <node name="translate_cell" channel="inline" format="flow/branch" model="">
 *     <node name="criterion" channel="inline" format="text/cybol-path" model="#empty"/>
 *     <node name="true" channel="inline" format="text/cybol-path" model=".logic.translate.db_to_wui.cell_default"/>
 *     <node name="false" channel="inline" format="text/cybol-path" model=".logic.translate.db_to_wui.cell"/>
 * </node>
 *
 * <!-- Call operation directly inline (not as cybol-path) and hand over properties as operation parametres. -->
 * <node name="process_model_as_compound_or_primitive" channel="inline" format="flow/branch" model="">
 *     <node name="criterion" channel="inline" format="text/cybol-path" model="#compound_flag"/>
 *     <node name="true" channel="inline" format="communicate/send" model="">
 *         <node name="channel" channel="inline" format="meta/channel" model="terminal"/>
 *         <node name="receiver" channel="inline" format="text/cybol-path" model=".stdout"/>
 *         <node name="language" channel="inline" format="meta/language" model="message/tui"/>
 *         <node name="format" channel="inline" format="meta/format" model="text/plain"/>
 *         <node name="message" channel="inline" format="text/plain" model="Do nothing, since this is a compound model."/>
 *     </node>
 *     <node name="false" channel="inline" format="modify/overwrite" model="">
 *         <node name="destination" channel="inline" format="text/cybol-path" model=".wui.index.body.content.(#wui_record_name).table.(#wui_field_name).model"/>
 *         <node name="source" channel="inline" format="text/cybol-path" model="#field_model_text"/>
 *     </node>
 * </node>
 *
 * Properties:
 *
 * - criterion (required) [text/cybol-path | logicvalue/boolean]: The flag defining which of the two models to execute.
 * - true (optional) [text/cybol-path | element/part]: The logic knowledge model to be executed if the criterion is true.
 * - false (optional) [text/cybol-path | element/part]: The logic knowledge model to be executed if the criterion is false.
 */
static wchar_t* BRANCH_FLOW_LOGIC_CYBOL_FORMAT = L"flow/branch";
static int* BRANCH_FLOW_LOGIC_CYBOL_FORMAT_COUNT = NUMBER_11_INTEGER_STATE_CYBOI_MODEL_ARRAY;

/**
 * The flow/loop logic cybol format.
 *
 * Description:
 *
 * Loops the programme flow endlessly, until the break flag is set.
 *
 * Java example:
 *
 * while (true) {
 *     if (flag != false) {
 *         break;
 *     }
 * }
 *
 * Examples:
 *
 * <!-- Read model from heap (knowledge tree). -->
 * <node name="print_numbers" channel="inline" format="flow/loop" model="">
 *     <node name="break" channel="inline" format="text/cybol-path" model=".break"/>
 *     <node name="model" channel="inline" format="text/cybol-path" model=".model"/>
 * </node>
 *
 * <!-- Read model from stack. -->
 * <node name="loop_rows" channel="inline" format="flow/loop" model="">
 *     <node name="break" channel="inline" format="text/cybol-path" model="#row_break"/>
 *     <node name="model" channel="inline" format="text/cybol-path" model="#row_model"/>
 * </node>
 *
 * <!-- Read model from file. -->
 * <node name="collect_voltages" channel="inline" format="flow/loop" model="">
 *     <node name="break" channel="inline" format="text/cybol-path" model=".settings.adc_break"/>
 *     <node name="model" channel="file" format="element/part" model="adc/collect_voltages.cybol"/>
 * </node>
 *
 * <!-- Translate data records of the german medical standard xdt. -->
 * <node name="loop_records" channel="inline" format="flow/loop" model="">
 *     <node name="break" channel="inline" format="path/knowledge" model=".temporary.xdt.record.break"/>
 *     <node name="model" channel="inline" format="path/knowledge" model=".logic.translate.record"/>
 * </node>
 *
 * <!-- Generate several webpages. -->
 * <node name="loop_page" channel="inline" format="flow/loop" model="">
 *     <node name="break" channel="inline" format="text/cybol-path" model="#page_break"/>
 *     <node name="model" channel="inline" format="text/cybol-path" model=".logic.generate.page"/>
 * </node>
 *
 * <!-- Call randomiser as model. -->
 * <node name="generate_mines_in_game_minesweeper" channel="inline" format="flow/loop" model="">
 *     <node name="break" channel="inline" format="text/cybol-path" model=".app.break"/>
 *     <node name="model" channel="inline" format="text/cybol-path" model=".app.logic.randomiser"/>
 * </node>
 *
 * Properties:
 *
 * - break (required) [text/cybol-path | logicvalue/boolean]: The break flag that causes the loop to be left.
 * - model (required) [text/cybol-path | element/part]: The logic knowledge model (function) to be executed repeatedly by the loop.
 */
static wchar_t* LOOP_FLOW_LOGIC_CYBOL_FORMAT = L"flow/loop";
static int* LOOP_FLOW_LOGIC_CYBOL_FORMAT_COUNT = NUMBER_9_INTEGER_STATE_CYBOI_MODEL_ARRAY;

/**
 * The flow/loop-each logic cybol format.
 *
 * Description:
 *
 * Iterates through all of the elements in the given collection.
 * The single elements may be accessed by index from within the model (function) being executed.
 *
 * This is somewhat similar to functional programming such as in the clojure programming language.
 * There are two kinds of behaviour that can be simulated in this way:
 * 1 "map" function (not container), other synonyms: "apply" or "consumer-accept-action"
 * 2 "reduce" function
 *
 * Java example:
 *
 * numbers.forEach( (n) -> { System.out.println(n); } );
 *
 * Clojure examples:
 *
 * 1 "map" function
 *
 * (map inc [1 2 3 4 5]) ;; => [2 3 4 5 6]
 * (map #(println %) [5 9 8 1])
 * (map #(str "Hello " % "!" ) ["Ford" "Arthur" "Tricia"])
 *
 * 2 "reduce" function
 *
 * (reduce + [1 2 3 4 5]) ;; => 15
 * (reduce conj #{} [:a :b :c]) ;; => #{:a :c :b}
 *    ;; Note that the output set #{} is unsorted, so that #{:a :c :b} appear in a different order.
 *
 * Reference:
 * https://clojuredocs.org/clojure.core/map
 *
 * Examples:
 *
 * <node name="iterate_over_elements" channel="inline" format="flow/loop-each" model="">
 *     <node name="collection" channel="inline" format="text/cybol-path" model=".collection"/>
 *     <node name="index" channel="inline" format="text/cybol-path" model=".index"/>
 *     <node name="model" channel="inline" format="text/cybol-path" model=".model"/>
 * </node>
 *
 * <!--
 *     This is a possible implementation of the model being called by the loop.
 *     It accesses the elements as destination, in order to write to them.
 *     This is the style of "map" (function, not container) in functional programming.
 * -->
 * <node name="increment_numbers" channel="inline" format="calculate/add" model="">
 *     <node name="result" channel="inline" format="text/cybol-path" model=".collection.[.index]"/>
 *     <node name="operand" channel="inline" format="number/integer" model="1"/>
 * </node>
 *
 * <!--
 *     This is a possible implementation of the model being called by the loop.
 *     It accesses the elements as source, in order to read from them.
 *     This is the style of "reduce" in functional programming.
 * -->
 * <node name="sum_up_numbers" channel="inline" format="calculate/add" model="">
 *     <node name="result" channel="inline" format="text/cybol-path" model=".sum"/>
 *     <node name="operand" channel="inline" format="text/cybol-path" model=".collection.[.index]"/>
 * </node>
 *
 * Properties:
 *
 * - collection (required) [text/cybol-path | element/part]: The collection containing the elements through which to iterate.
 * - index (required) [text/cybol-path]: The index of the current element.
 * - model (required) [text/cybol-path | element/part]: The logic knowledge model (function) to be executed repeatedly for each element.
 */
static wchar_t* LOOP_EACH_FLOW_LOGIC_CYBOL_FORMAT = L"flow/loop-each";
static int* LOOP_EACH_FLOW_LOGIC_CYBOL_FORMAT_COUNT = NUMBER_14_INTEGER_STATE_CYBOI_MODEL_ARRAY;

/**
 * The flow/loop-for logic cybol format.
 *
 * Description:
 *
 * Loops the programme flow from the given start to the given end in the given steps.
 *
 * The comparison of index and end value is exclusive, that is it does not
 * consider the end value in the executed model since the loop is left before:
 * - positive step: if (index >= endvalue) then break
 * - negative step: if (index <= endvalue) then break
 *
 * If the property "step" contains a negative value, then the loop is running downwards.
 *
 * Java example:
 *
 * for (int i = 0; i < 10; i += 2) {
 *     System.out.println("Hello");
 * }
 *
 * Examples:
 *
 * <node name="print_numbers" channel="inline" format="flow/loop-for" model="">
 *     <node name="start" channel="inline" format="number/integer" model="0"/>
 *     <node name="end" channel="inline" format="number/integer" model="10"/>
 *     <node name="step" channel="inline" format="number/integer" model="2"/>
 *     <node name="index" channel="inline" format="text/cybol-path" model=".index"/>
 *     <node name="model" channel="inline" format="text/cybol-path" model=".model"/>
 * </node>
 *
 * <node name="run_upwards" channel="inline" format="flow/loop-for" model="">
 *     <node name="start" channel="inline" format="number/integer" model="20"/>
 *     <node name="end" channel="inline" format="number/integer" model="37"/>
 *     <node name="step" channel="inline" format="number/integer" model="2"/>
 *     <node name="index" channel="inline" format="text/cybol-path" model=".index"/>
 *     <node name="model" channel="inline" format="text/cybol-path" model=".model"/>
 * </node>
 *
 * <node name="run_downwards" channel="inline" format="flow/loop-for" model="">
 *     <node name="start" channel="inline" format="number/integer" model="17"/>
 *     <node name="end" channel="inline" format="number/integer" model="-4"/>
 *     <node name="step" channel="inline" format="number/integer" model="-3"/>
 *     <node name="index" channel="inline" format="text/cybol-path" model=".index"/>
 *     <node name="model" channel="inline" format="text/cybol-path" model=".model"/>
 * </node>
 *
 * Properties:
 *
 * - start (required) [text/cybol-path | number/integer]: The start value that gets assigned to the index before the looping begins.
 * - end (required) [text/cybol-path | number/integer]: The end value to be reached by the index, in which case the looping ends. It is exclusive and is not considered itself.
 * - step (required) [text/cybol-path | number/integer]: The step width for counting the index up or down. It may be negative.
 * - index (required) [text/cybol-path]: The index of the current element.
 * - model (required) [text/cybol-path | element/part]: The logic knowledge model (function) to be executed repeatedly by the loop.
 */
static wchar_t* LOOP_FOR_FLOW_LOGIC_CYBOL_FORMAT = L"flow/loop-for";
static int* LOOP_FOR_FLOW_LOGIC_CYBOL_FORMAT_COUNT = NUMBER_13_INTEGER_STATE_CYBOI_MODEL_ARRAY;

/**
 * The flow/loop-times logic cybol format.
 *
 * Description:
 *
 * Loops the programme flow as often as specified by the times property.
 *
 * Clojure example:
 *
 * (dotimes [_ 10]
 *     (println "Hello"))
 *
 * Examples:
 *
 * <node name="print_numbers" channel="inline" format="flow/loop-times" model="">
 *     <node name="times" channel="inline" format="number/integer" model="10"/>
 *     <node name="model" channel="inline" format="text/cybol-path" model=".model"/>
 * </node>
 *
 * Properties:
 *
 * - times (required) [text/cybol-path | number/integer]: The number of cycles the loop should run.
 * - model (required) [text/cybol-path | element/part]: The logic knowledge model (function) to be executed repeatedly by the loop.
 */
static wchar_t* LOOP_TIMES_FLOW_LOGIC_CYBOL_FORMAT = L"flow/loop-times";
static int* LOOP_TIMES_FLOW_LOGIC_CYBOL_FORMAT_COUNT = NUMBER_15_INTEGER_STATE_CYBOI_MODEL_ARRAY;

/**
 * The flow/sequence logic cybol format.
 *
 * Description:
 *
 * Executes the given programme flow as sequence.
 *
 * This encapsulating operation "flow/sequence" is provided here
 * to have all three constructs of structural programming implemented
 * together, which are "sequence", "loop" and "branch".
 *
 * However, two kinds of executing logic models are possible, either:
 * - standalone directly or
 * - via the encapsulating operation "flow/sequence".
 *
 * Examples:
 *
 * <node name="start_player_input" channel="inline" format="flow/sequence" model="">
 *     <node name="model" channel="inline" format="text/cybol-path" model=".app.setrandom"/>
 * </node>
 *
 * <node name="send_index" channel="inline" format="flow/sequence" model="">
 *     <node name="model" channel="inline" format="text/cybol-path" model=".logic.send.index"/>
 * </node>
 *
 * <node name="execute_action" channel="inline" format="flow/sequence" model="">
 *     <node name="model" channel="inline" format="text/cybol-path" model="{.path}"/>
 * </node>
 *
 * <node name="translate_uri_to_wui" channel="inline" format="flow/sequence" model="">
 *     <node name="model" channel="inline" format="text/cybol-path" model=".logic.translate.uri_to_wui.column"/>
 * </node>
 *
 * <!-- Direct execution of logic model without encapsulating operation "flow/sequence". -->
 * <node name="navigate_to_week" channel="inline" format="text/cybol-path" model=".logic.navigate.week"/>
 *
 * Properties:
 *
 * - model (required) [text/cybol-path | element/part]: The logic knowledge model to be executed as sequence.
 */
static wchar_t* SEQUENCE_FLOW_LOGIC_CYBOL_FORMAT = L"flow/sequence";
static int* SEQUENCE_FLOW_LOGIC_CYBOL_FORMAT_COUNT = NUMBER_13_INTEGER_STATE_CYBOI_MODEL_ARRAY;

/* FLOW_LOGIC_CYBOL_FORMAT_CONSTANT_HEADER */
#endif
